@@ -24,8 +24,6 @@ class JSONConfigParser(ConfigParserBase):
         if preparsed_content["content"] == "":
             raise ValueError("Cannot parse empty configuration")
 
-        with self.statsd.get_timer().time("parse_time"):
-            config = json.loads(preparsed_content["content"])
+        config = json.loads(preparsed_content["content"])
 
-        self.statsd.get_counter().increment("parse")
         return config
