@@ -33,7 +33,10 @@ class RedisStore(ConfigStoreBase):
 
         super().retrieve(name)
 
-        config = self.redis.get("saas:instance:" + name)
+        timer = self.statsd.get_timer("config_database_read_duration")
+        with timer.time():
+            config = self.redis.get("saas:instance:" + name)
+        self.statsd.get_counter("config_database_read_number").increment()
 
         if config is None:
             raise ConfigurationNotFound
