@@ -2,17 +2,33 @@
 #
 # SPDX-License-Identifier: EUPL-1.2
 
-import wrapt
+import functools
+from .cqrs import EventService
+from .cqrs.events import EntityChange
 from abc import ABC, abstractmethod
 from datetime import datetime
 from enum import Enum
 from minty import exceptions
 from minty.object import Field, IntrospectableObject
-from typing import Iterable, List
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Generic,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    TypeVar,
+)
 from uuid import UUID
 
 
 class EntityBase(ABC):
+    _event_service: Optional[EventService]
+    _changes: List[EntityChange]
+    _entity_data: Dict[str, Any]
+
     @property
     @abstractmethod
     def entity_id(self):
@@ -51,7 +67,7 @@ class EntityBase(ABC):
     def __setattr__(self, attr, value):
         try:
             old_value = self.__getattribute__(attr)
-            change = {
+            change: EntityChange = {
                 "key": attr,
                 "old_value": _reflect(old_value),
                 "new_value": _reflect(value),
@@ -113,6 +129,75 @@ class ValueObject(IntrospectableObject):
     pass
 
 
+def _event_decorator(
+    event_name: str,
+    extra_fields: Optional[List[str]] = None,
+):
+    def _build_decorator(wrapped: Callable[..., Any]):
+        @functools.wraps(wrapped)
+        def wrapper(*args, **kwargs):
+            instance = args[0]
+
+            # Clear the changelog when we are not a class method
+            if isinstance(instance, Entity):
+                instance.entity_changelog = []
+
+            rv = wrapped(*args, **kwargs)
+
+            changelog: List[EntityChange]
+
+            if isinstance(instance, Entity):
+                # When used with a regular method, use the regular
+                changelog = instance.entity_changelog
+                wrapped_entity = instance
+            else:
+                if not isinstance(rv, Entity):
+                    raise exceptions.ConfigurationConflict(
+                        "Return value of class method is not an entity."
+                    )
+
+                # When called as a class method (to create a new entity)
+                # everything is considered a change
+                dict_of_changes = rv.entity_dict()
+
+                changelog = []
+                for field in dict_of_changes.keys():
+                    value = dict_of_changes[field]
+
+                    changelog.append(
+                        {
+                            "key": field,
+                            "old_value": None,
+                            "new_value": _reflect(value),
+                        }
+                    )
+
+                wrapped_entity = rv
+
+            if extra_fields is not None:
+                wrapped_entity.capture_field_values(fields=extra_fields)
+
+            if wrapped_entity._event_service:
+                wrapped_entity._event_service.log_event(
+                    entity_type=wrapped_entity.__class__.__name__,
+                    entity_id=str(wrapped_entity.entity_id),
+                    event_name=event_name,
+                    changes=changelog,
+                    entity_data=wrapped_entity.entity_data,
+                )
+            else:
+                raise exceptions.ConfigurationConflict(
+                    "Wrapped entity does not have _event_service attribute. "
+                    "Possible cause: repository did not inject it."
+                )
+
+            return wrapped_entity
+
+        return wrapper
+
+    return _build_decorator
+
+
 class Entity(IntrospectableObject):
     """
     Pydantic based Entity object
@@ -159,7 +244,7 @@ class Entity(IntrospectableObject):
     entity_changelog: list = []
     entity_data: dict = {}
 
-    _event_service = None
+    _event_service: Optional[EventService] = None
 
     def __init__(self, **kwargs):
         """Initialized an Entity object
@@ -221,74 +306,42 @@ class Entity(IntrospectableObject):
             self.entity_data[field] = _reflect(value)
 
     @staticmethod
-    def event(name: str, extra_fields: List = None):
-        """New event decorator to capture entity changes and publish events to
-        the event_service.
-
-        :param name: Name of event to publish
-        :type name: str
-        :return: wrapped_entity
-        :rtype: func
+    def event(name: str, extra_fields: Optional[List[str]] = None):
         """
+        Decorator to defined events on entities.
 
-        @wrapt.decorator
-        def wrapper(wrapped, instance, args, kwargs):
-            # Clear the changelog when we are not a classmethod
-            if isinstance(instance, Entity):
-                instance.entity_changelog = []
+        Captures entity changes and saves them to the event_service as events.
 
-            rv = wrapped(*args, **kwargs)
+        If specified, the contents of fields named in `extra_fields` are
+        also included in the event.
 
-            # Check if we are a classmethod instantiating this class. If so,
-            # get the changelog from the return value
-            changelog = []
-            wrapped_entity = None
-            if isinstance(instance, Entity):
-                changelog = instance.entity_changelog
-                wrapped_entity = instance
-            elif isinstance(rv, Entity):
-                # Classmethod, everything is a change
-                dict_of_changes = rv.entity_dict()
-                for field in dict_of_changes.keys():
-                    value = dict_of_changes[field]
+        When applying this to a class method, the `@classmethod` decorator
+        should be the topmost one:
 
-                    changelog.append(
-                        {
-                            "key": field,
-                            "old_value": None,
-                            "new_value": _reflect(value),
-                        }
-                    )
-
-                wrapped_entity = rv
-
-            if extra_fields is not None:
-                wrapped_entity.capture_field_values(fields=extra_fields)
-
-            try:
-                wrapped_entity._event_service.log_event(
-                    entity_type=wrapped_entity.__class__.__name__,
-                    entity_id=str(wrapped_entity.entity_id),
-                    event_name=name,
-                    changes=changelog,
-                    entity_data=wrapped_entity.entity_data,
-                )
-            except AttributeError as e:
-                raise exceptions.ConfigurationConflict(
-                    "Wrapped entity does not have _event_service attribute. "
-                    "Possible cause: class method did not return an entity."
-                ) from e
-
-            return wrapped_entity
-
-        return wrapper
+        ```
+            @classmethod
+            @Entity.event("EventNamed")
+            def event_name(cls):
+                pass
+        ```
+        """
+        return _event_decorator(event_name=name, extra_fields=extra_fields)
 
 
-class EntityCollection:
+E = TypeVar("E", bound=Entity)
+
+
+class EntityCollection(Generic[E]):
     """Multiple entities."""
 
-    def __init__(self, entities: Iterable[Entity]):
-        self.entities = entities
+    total_results: Optional[int]
+    entities: Iterable[E]
 
-    def __iter__(self):
+    def __init__(
+        self, entities: Iterable[E], total_results: Optional[int] = None
+    ):
+        self.entities = entities
+        self.total_results = total_results
+
+    def __iter__(self) -> Iterator[E]:
         return iter(self.entities)
